"""Tests for the raster utilities module."""
import doctest
import unittest

from aatoolbox.utils import raster

suite = unittest.TestSuite()
suite.addTest(doctest.DocTestSuite(raster))
unittest.TextTestRunner().run(suite)



"""Download and manipulate COD administrative boundaries."""
from pathlib import Path

import geopandas as gpd

from aatoolbox.datasources.datasource import DataSource
from aatoolbox.utils.hdx_api import load_dataset_from_hdx
from aatoolbox.utils.io import check_file_existence

_MODULE_BASENAME = "cod_ab"


class CodAB(DataSource):
    """
    Work with COD AB (administrative boundaries).

    Parameters
    ----------
    iso3 : (str)
        country iso3
    """

    def __init__(self, iso3: str):
        super().__init__(
            iso3=iso3, module_base_dir=_MODULE_BASENAME, is_public=True
        )
        self._raw_filepath = (
            self._raw_base_dir / f"{self._iso3}_{_MODULE_BASENAME}.shp.zip"
        )

    def download(
        self, hdx_address: str, hdx_dataset_name: str, clobber: bool = False
    ) -> Path:
        """
        Download COD AB file from HDX.

        Parameters
        ----------
        hdx_address: str
            URL suffix of dataset page on HDX
        hdx_dataset_name: str
            Name of dataset on HDX
        clobber : bool, default = False
            If True, overwrites existing COD AB files

        Returns
        -------
        The downloaded filepath

        Examples
        --------
        >>> from aatoolbox.datasources.codab.codab import CodAB
        >>> # Download COD administrative boundaries for Nepal
        >>> codab = CodAB("npl")
        >>> npl_cod_shapefile = codab.download(
        ...    hdx_address="administrative-bounadries-of-nepal",
        ...    hdx_dataset_name="npl_admbnda_nd_20201117_SHP.zip")
        """
        return self._download(
            filepath=self._raw_filepath,
            hdx_address=hdx_address,
            hdx_dataset_name=hdx_dataset_name,
            clobber=clobber,
        )

    @staticmethod
    @check_file_existence
    def _download(
        filepath: Path, hdx_address: str, hdx_dataset_name: str, clobber: bool
    ):
        return load_dataset_from_hdx(
            hdx_address=hdx_address,
            hdx_dataset_name=hdx_dataset_name,
            output_filepath=filepath,
        )

    def load_admin_layer(self, layer_name: str) -> gpd.GeoDataFrame:
        """
        Get an admin level by layer name.

        Parameters
        ----------
        layer_name: str
            The admin layer name

        Returns
        -------
        geopandas dataframe with COD AB admin information

        Examples
        --------
        >>> from aatoolbox.datasources.codab.codab import CodAB
        >>> # Retrieve admin 0 boundaries for Nepal
        >>> codab = CodAB("npl")
        >>> npl_admin0 = codab.load_admin_layer(
        ...     layer_name="npl_admbnda_adm2_20201117.shp")
        """
        return gpd.read_file(f"zip:///{self._raw_filepath / layer_name}")



"""Methods pertaining to the COD administrative boundaries.

`Common Operational Datasets <https://cod.unocha.org>`
(CODs) are definitive reference datasets governed by OCHA Field Information
Section (FIS) and designed
to support decision making during a humanitarian response.
The Administrative Boundary (AB) CODs are geospatial datasets that
delineate a country's borders and internal regions.
A key feature of the COD AB datasets are P-codes, which are unique
alphanumeric identifiers for each geographic region.
"""



"""aa-toolbox: Toolbox for Anticipatory Action.

The ``aatoolbox`` library contains modules designed to assist with
downloading and processing the data required for disaster-related
anticipatory action.
"""
from ._version import version as __version__  # noqa: F401



"""
FEWS NET processing.

Download and save the data provided by FEWS NET as provided on .
"""

import logging
import zipfile
from datetime import date, datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, Union

from typing_extensions import Literal

from aatoolbox.utils.io import download_url, unzip

logger = logging.getLogger(__name__)
_DATE_FORMAT = "%Y-%m-%d"
_BASE_URL_COUNTRY = (
    "https://fdw.fews.net/api/ipcpackage/"
    "?country_code={iso2}&collection_date={YYYY}-{MM}-01"
)
_BASE_URL_REGION = (
    "https://fews.net/data_portal_download/download"
    "?data_file_path=http://shapefiles.fews.net.s3.amazonaws.com/"
    "HFIC/{region_code}/{region_name}{YYYY}{MM}.zip"
)

_VALID_REGION_NAMES = [
    "caribbean-central-america",
    "central-asia",
    "east-africa",
    "southern-africa",
    "west-africa",
]
_VALID_REGION_CODES = ["LAC", "EA", "CA", "SA", "WA"]


def download_fewsnet(
    date_pub: Union[date, str],
    iso2: str,
    region_name: Literal[
        "caribbean-central-america",
        "central-asia",
        "east-africa",
        "southern-africa",
        "west-africa",
    ],
    region_code: Literal["LAC", "EA", "CA", "SA", "WA"],
    output_dir: Union[Path, str],
    use_cache=True,
) -> Path:
    """
    Retrieve the raw fewsnet data.

    Depending on the region and date, this data is published per region or
    per country. This function retrieves the country data
    if it exists, and else the regional data for `date_pub` and `iso2`.

    Parameters
    ----------
    date_pub : date or str
        date for which the data should be downloaded
        only the year and month part are used
        this commonly refers to the month of the Current Situation period
        if str, it should be in ISO 8601 format, e.g. '2021-10-01'
    iso2 : str
        iso2 code of the country of interest.
        See https://fews.net/ for a list of countries that are
        included by FewsNet
    region_name : {'caribbean-central-america', 'central-asia',
        'east-africa', 'southern-africa', 'west-africa'}
        name of the region to which the `iso2` belongs.
        Only regions for which FewsNet provides data can be given as input
    region_code : {'LAC','EA','CA','SA','WA'}
        code that refers to the `region_name`
        These are defined by FewsNet
    output_dir : Path or str
        path to dir to which the data should be written
    use_cache : bool, default True
        if True, don't download if output_dir already exists

    Examples
    --------
    >>> download_fewsnet(date_pub="2020-10-01",iso2="et",
    ... region_code="east-africa", region_name="EA",
    ... output_dir="tmp",use_cache=False)
    """
    if region_name not in _VALID_REGION_NAMES:
        raise ValueError(f"Invalid region name {region_name}")
    if region_code not in _VALID_REGION_CODES:
        raise ValueError(f"Invalid region code {region_code}")

    # convert to datetime if str
    if not isinstance(date_pub, date):
        date_pub = datetime.strptime(date_pub, _DATE_FORMAT)
    # convert to path object if str
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)

    # we prefer the country data as this more nicely structured
    # thus first check if that is available
    country_path = _download_fewsnet_country(
        date_pub=date_pub,
        iso2=iso2,
        output_dir=output_dir,
        use_cache=use_cache,
    )
    if country_path is not None:
        return country_path
    else:
        region_path = _download_fewsnet_region(
            date_pub=date_pub,
            region_name=region_name,
            region_code=region_code,
            output_dir=output_dir,
            use_cache=use_cache,
        )
    if region_path is not None:
        return region_path
    else:
        raise RuntimeError(f"No data found for {date_pub.strftime('%Y-%m')}")


def _download_fewsnet_country(
    date_pub: date,
    iso2: str,
    output_dir: Path,
    use_cache: bool,
) -> Optional[Path]:
    """
    Download fewsnet data that covers the iso2 country.

    Parameters
    ----------
    date_pub : date
        date for which the data should be downloaded
        only the year and month part are used
        this commonly refers to the month of the Current Situation period
    iso2 : str
        iso2 code for which the data should be downloaded
    output_dir: Path
        path of dir to which the zip content should be written
    use_cache: bool
        if True, don't download if output_dir already exists

    Returns
    -------
    country_data : str
        if data found return the output_dir, else return None
    """
    url_country_date = _BASE_URL_COUNTRY.format(
        iso2=iso2.upper(), YYYY=date_pub.year, MM=date_pub.month
    )

    # filenames have upper iso2, so use that for dirs as well
    output_dir_country = (
        output_dir / f"{iso2.upper()}{date_pub.strftime('%Y%m')}"
    )
    zip_filename = f"{output_dir_country.name}.zip"
    if not output_dir_country.exists() or use_cache is False:
        country_data = _download_zip(
            url=url_country_date,
            zip_filename=zip_filename,
            output_dir=output_dir_country,
        )
    else:
        country_data = output_dir_country
    return country_data


def _download_fewsnet_region(
    date_pub: date,
    region_name: str,
    region_code: str,
    output_dir: Path,
    use_cache: bool,
) -> Optional[Path]:
    """
    Download fewsnet data that covers `region_name`.

    Parameters
    ----------
    date_pub : date
        date for which the data should be downloaded
        only the year and month part are used
        this commonly refers to the month of the Current Situation period
    region_name : str
        name of the region to which the `iso2` belongs,
        e.g. "east-africa"
    region_code : str
        code that refers to the `region_name,
        e.g. "EA"
    output_dir : Path
        path to dir to which the data should be written
    use_cache : bool
        if True, don't download if output_dir already exists

    Returns
    -------
    region_data : str
        If region data exists, return the saved dir else return None
    """
    url_region_date = _BASE_URL_REGION.format(
        region_code=region_code.upper(),
        region_name=region_name,
        YYYY=date_pub.year,
        MM=date_pub.month,
    )

    output_dir_region = (
        output_dir / f"{region_code.upper()}{date_pub.strftime('%Y%m')}"
    )
    zip_filename_region = f"{output_dir_region.name}.zip"
    if not output_dir_region.exists() or use_cache is False:
        region_data = _download_zip(
            url=url_region_date,
            zip_filename=zip_filename_region,
            output_dir=output_dir_region,
        )
    else:
        region_data = output_dir_region
    return region_data


def _download_zip(
    url: str,
    zip_filename: str,
    output_dir: Path,
) -> Optional[Path]:
    """
    Download and unzip the file at the url.

    Parameters
    ----------
    url : str
        url that contains the zip file to be downloaded
    zip_filename : str
        name of the zipfile
    output_dir : Path
        path of dir to which the zip content should be written

    Returns
    -------
    None if no valid file, else output_dir

    """
    # create tempdir to write zipfile to
    with TemporaryDirectory() as temp_dir:
        zip_path = Path(temp_dir) / zip_filename
        download_url(url=url, save_path=zip_path)
        logger.info(f"Downloaded {url} to {zip_path}")

        try:
            unzip(zip_file_path=zip_path, save_dir=output_dir)
            logger.debug(f"Unzipped to {output_dir}")

        except zipfile.BadZipFile:
            # indicates that the url returned something that wasn't a
            # zip, happens often and indicates data for the given
            # country - date is not available
            logger.debug(
                f"No zip data returned from url {url} "
                f"check that the area and date exist."
            )
            return None

    return output_dir



"""Country configuration setting base class."""
from pathlib import Path
from typing import Optional

from pydantic import BaseModel

from aatoolbox.utils.io import parse_yaml


class CodABConfig(BaseModel):
    """COD AB configuration.

    Parameters
    ----------
    hdx_address: str
        The page where the COD AB dataset resides on on HDX. Can be found
        by taking the portion of the url after ``data.humdata.org/dataset/``
    hdx_dataset_name: str
        COD AB dataset name on HDX. Can be found by taking the filename as it
        appears on the dataset page.
    layer_base_name: str
        The base name of the different admin layers, that presumably only
        change by a single custom_layer_number depending on the level. Should
        contain {admin_level} in place of the custom_layer_number.
    admin_level_max: int
        The maximum admin level available in the shapefile.
    custom_layer_names: list, optional
        Any additional layer names that don't fit into the admin level paradigm
    """

    hdx_address: str
    hdx_dataset_name: str
    layer_base_name: str  # TODO: validate that it has {admin_level}
    admin_level_max: int
    custom_layer_names: Optional[list]


class CountryConfig(BaseModel):
    """Country configuration."""

    iso3: str
    codab: CodABConfig


def get_country_config(iso3: str) -> CountryConfig:
    """
    Return a country configuration object.

    Parameters
    ----------
    iso3 : str
        Country ISO3, must e exactly 3 characters long

    Returns
    -------
    CountryConfig instance
    """
    # TODO: validate iso3
    parameters = parse_yaml(
        Path(__file__).parent.resolve() / f"countries/{iso3}.yaml"
    )
    return CountryConfig(**parameters)



"""Test COD AB methods as called from the pipeline."""
from pathlib import Path

import pytest
from conftest import FAKE_AA_DATA_DIR, ISO3

from aatoolbox.datasources.codab.codab import _MODULE_BASENAME


@pytest.fixture(autouse=True)
def downloader(mocker):
    """Mock the HDX download function."""
    return mocker.patch(
        "aatoolbox.datasources.codab.codab.load_dataset_from_hdx"
    )


@pytest.fixture(autouse=True)
def gpd_read_file(mocker):
    """Mock GeoPandas file reading function."""
    return mocker.patch("aatoolbox.datasources.codab.codab.gpd.read_file")


def test_codab_download(pipeline, downloader):
    """Test that load_codab calls the HDX API to download."""
    pipeline.load_codab(admin_level=2)
    downloader.assert_called_with(
        hdx_address=pipeline._config.codab.hdx_address,
        hdx_dataset_name=pipeline._config.codab.hdx_dataset_name,
        output_filepath=Path(FAKE_AA_DATA_DIR)
        / f"public/raw/{ISO3}/{_MODULE_BASENAME}/"
        f"{ISO3}_{_MODULE_BASENAME}.shp.zip",
    )


def test_codab_load_admin_level(pipeline, gpd_read_file):
    """Test that load_codab retrieves expected file and layer name."""
    admin_level = 2
    expected_layer_name = pipeline._config.codab.layer_base_name.format(
        admin_level=admin_level
    )

    pipeline.load_codab(admin_level=admin_level)

    gpd_read_file.assert_called_with(
        f"zip:///{FAKE_AA_DATA_DIR}/public/raw/{ISO3}/{_MODULE_BASENAME}/"
        f"{ISO3}_{_MODULE_BASENAME}.shp.zip/{expected_layer_name}"
    )


def test_codab_too_high_admin_level(pipeline):
    """Test raised error when too high admin level requested."""
    with pytest.raises(AttributeError):
        pipeline.load_codab(admin_level=10)


def test_codab_custom(pipeline, gpd_read_file):
    """Test that load_codab_custom retrieves expected file and layer name."""
    custom_layer_number = 1
    custom_layer_name_list = ["custom_layer_A", "custom_layer_B"]
    pipeline._config.codab.custom_layer_names = custom_layer_name_list
    pipeline.load_codab_custom(custom_layer_number)
    gpd_read_file.assert_called_with(
        f"zip:///{FAKE_AA_DATA_DIR}/public/raw/{ISO3}/{_MODULE_BASENAME}/"
        f"{ISO3}_{_MODULE_BASENAME}.shp.zip/"
        f"{custom_layer_name_list[custom_layer_number]}"
    )


def test_codab_custom_missing(pipeline, gpd_read_file):
    """Test raised error when custom COD AB missing."""
    with pytest.raises(AttributeError):
        pipeline.load_codab_custom(0)



"""Tests for the CDS GeoBoundingBox module."""
from geopandas import GeoSeries
from shapely.geometry import Polygon

from aatoolbox.utils.geoboundingbox import GeoBoundingBox


def test_geoboundingbox_round_coords():
    """Test that coordinates are correctly rounded and offset."""
    geobb = GeoBoundingBox(north=1.05, south=-2.2, east=3.6, west=-4)
    geobb.round_boundingbox_coords(offset_val=0.4)
    assert geobb.north == 2.4
    assert geobb.south == -3.4
    assert geobb.east == 4.4
    assert geobb.west == -4.4


def test_geoboundingbox_filename():
    """Test that correct file string is returned."""
    geobb = GeoBoundingBox(north=1, south=-2, east=3, west=-4)
    assert geobb.get_filename_repr() == "Np1Sm2Ep3Wm4"


def test_geoboundingbox_from_shape():
    """Test that getting the geobb from a polygon works as expected."""
    n, s, e, w = (1, -2, 3, -4)
    shape = GeoSeries([Polygon([(e, n), (e, s), (w, s), (w, n)])])
    geobb = GeoBoundingBox.from_shape(shape)
    assert geobb.north == n
    assert geobb.south == s
    assert geobb.east == e
    assert geobb.west == w



"""
Functionality to retrieve and modify boundary coordinates.

It is possible to create an ``GeoBoundingBox`` object either from
north, south, east, west coordinates,
or from a shapefile that has been read in with geopandas.
"""
from typing import Union

import geopandas as gpd
import numpy as np


class GeoBoundingBox:
    """Create an object containing the bounds of an area.

    Parameters
    ----------
    north : float
        The northern latitude boundary of the area (degrees)
    south : float
        The southern latitude boundary of the area (degrees)
    east : float
        The easternmost longitude boundary of the area (degrees)
    west : float
        The westernmost longitude boundary of the area (degrees)
    """

    def __init__(self, north: float, south: float, east: float, west: float):
        self.north = north
        self.south = south
        self.east = east
        self.west = west

    def __repr__(self):
        """Print bounding box string."""
        return (
            f"N: {self.north}\nS: {self.south}\n"
            f"E: {self.east}\nW: {self.west}"
        )

    @classmethod
    def from_shape(cls, shape: Union[gpd.GeoSeries, gpd.GeoDataFrame]):
        """
        Create `GeoBoundingBox` from a geopandas object.

        Parameters
        ----------
        shape : geopandas.GeoSeries, geopandas.GeoDataFrame
            A shape whose bounds will be retrieved

        Examples
        --------
        >>> import geopandas as gpd
        >>> df_admin_boundaries = gpd.read_file("admin0_boundaries.gpkg")
        >>> geobb = GeoBoundingBox.from_shape(df_admin_boundaries)
        """
        return cls(
            north=shape.total_bounds[3],
            south=shape.total_bounds[1],
            east=shape.total_bounds[2],
            west=shape.total_bounds[0],
        )

    def round_boundingbox_coords(
        self,
        offset_val: float = 0.0,
        round_val: Union[int, float] = 1,
    ):
        """
        Round the bounding box coordinates.

        Rounding is always done outside the original bounding box.
        I.e. the resulting bounding box is always equal or larger
        than the original bounding box.

        Parameters
        ----------
        offset_val : float, default = 0.0
            Offset the coordinates by this factor.
        round_val : int, default = 1
            The decimal to round to. If 1, round to integers
        """
        for direction in ["north", "west", "south", "east"]:
            coord = getattr(self, direction)
            if direction in ("north", "east"):
                function = np.ceil.__call__  # needed for mypy
                offset_factor = 1
            elif direction in ("south", "west"):
                function = np.floor.__call__  # needed for mypy
                offset_factor = -1
            rounded_coord = (
                function(coord / round_val) * round_val
                + offset_factor * offset_val
            )
            setattr(self, direction, rounded_coord)  # noqa: FKA01

    def get_filename_repr(self, p: int = 0) -> str:
        """
        Get succinct boundary representation for usage in filenames.

        Parameters
        ----------
        p : int, default = 0
            Precision, i.e. number of decimal places to round to. Default is
            0 for ints.

        Returns
        -------
        String containing N, S, E and W coordinates.
        """

        def _str_format(coord):
            """Add m indicating minus value and p indicating positive value."""
            if coord < 0:
                return f"m{abs(coord):.{p}f}"
            else:
                return f"p{coord:.{p}f}"

        return (
            f"N{_str_format(self.north)}S{_str_format(self.south)}"
            f"E{_str_format(self.east)}W{_str_format(self.west)}"
        )



"""Base class for aatoolbox data source."""
from pathlib import Path

from aatoolbox.config.pathconfig import PathConfig


class DataSource:
    """
    Base class object that contains path convenience functions.

    Parameters
    ----------
    iso3: str
        Country ISO3
    module_base_dir : str
        Module directory name (usually correspond to data source)
    is_public: bool, default = False
        Whether the dataset is public or private. Determines top-level
        directory structure.
    """

    def __init__(
        self, iso3: str, module_base_dir: str, is_public: bool = False
    ):

        self._iso3 = iso3
        self._module_base_dir = module_base_dir
        self._path_config = PathConfig()
        self._raw_base_dir = self._get_base_dir(
            is_public=is_public, is_raw=True
        )
        self._processed_base_dir = self._get_base_dir(
            is_public=is_public, is_raw=False
        )

    def _get_base_dir(self, is_public: bool, is_raw: bool) -> Path:
        permission_dir = (
            self._path_config.public
            if is_public
            else self._path_config.private
        )
        state_dir = (
            self._path_config.raw if is_raw else self._path_config.processed
        )
        return (
            self._path_config.base_path
            / permission_dir
            / state_dir
            / self._iso3
            / self._module_base_dir
        )



"""Function for I/O."""
import logging
import zipfile
from pathlib import Path
from typing import Any, Callable, TypeVar, Union, cast

import requests
import yaml

logger = logging.getLogger(__name__)

# For typing the decorator
F = TypeVar("F", bound=Callable[..., Any])


def download_url(
    url: str,
    save_path: Path,
    chunk_size: int = 2048,
):
    """
    Download the file located at `url` to `save_path`.

    Parameters
    ----------
    url : str
        url that contains the file to be downloaded
    save_path : Path
        path to the location the file should be saved
    chunk_size : int
        number of bytes to save at once
    """
    save_path.parent.mkdir(exist_ok=True, parents=True)
    # Remove file if already exists
    save_path.unlink(missing_ok=True)

    # use a session and chunk_size to prevent
    # crashing when downloading large files while
    # not loosing too much speed
    session = requests.Session()
    r = session.get(url, stream=True)
    r.raise_for_status()
    with save_path.open("wb") as fd:
        for chunk in r.iter_content(chunk_size=chunk_size):
            fd.write(chunk)


def unzip(
    zip_file_path: Path,
    save_dir: Path,
):
    """
    Unzip a file.

    Parameters
    ----------
    zip_file_path : Path
        path to the location the zip file is saved
    save_dir : Path
        dir path to which the content of the zip
        file should be saved
    """
    with zipfile.ZipFile(file=zip_file_path, mode="r") as zip_ref:
        zip_ref.extractall(save_dir)


def parse_yaml(filename: Union[str, Path]) -> dict:
    """
    Read in a yaml file.

    Parameters
    ----------
    filename : str, Path
    The full filepath of the YAML file

    Returns
    -------
    A dictionary with the YAML file contents
    """
    with open(file=filename, mode="r") as stream:
        config = yaml.safe_load(stream)
    return config


def check_file_existence(func: F) -> F:
    """
    Don't overwrite existing data.

    Avoid recreating data if it already exists and if clobber not
    toggled by user. Can be used to wrap a function whose first parameter
    is a filepath.

    Parameters
    ----------
    func : function
        The function to wrap. The first parameter of this function must
        be the filepath (of type Path), and it can also have an optional
        "clobber" boolean keyword parameter.

    Returns
    -------
    If filepath exists, returns filepath. Otherwise, returns the result of
    the decorated function.

    """

    def wrapper(filepath: Path, *args, **kwargs):
        if filepath.exists() and not kwargs.get("clobber", False):
            logger.info(
                f"File {filepath} exists and clobber set to False, "
                f"using existing files"
            )
            return filepath
        return func(filepath, *args, **kwargs)

    return cast(F, wrapper)



"""Fixtures for all pipeline-related tests."""
import pytest

from aatoolbox.pipeline import Pipeline
from aatoolbox.utils.io import parse_yaml

CONFIG_FILE = "tests/pipeline/fake_config.yaml"
FAKE_AA_DATA_DIR = "fake_aa_dir"
ISO3 = "abc"


@pytest.fixture(scope="session", autouse=True)
def mock_aa_data_dir(session_mocker):
    """Mock out the AA_DATA_DIR environment variable."""
    session_mocker.patch.dict(
        "aatoolbox.config.pathconfig.os.environ",
        {"AA_DATA_DIR": FAKE_AA_DATA_DIR},
    )


@pytest.fixture
def pipeline(mocker):
    """Fixture for pipeline with test config params."""
    mocker.patch(
        "aatoolbox.config.countryconfig.parse_yaml",
        return_value=parse_yaml(CONFIG_FILE),
    )
    return Pipeline(iso3_unvalidated=ISO3)



"""Pipeline initializer."""
import geopandas as gpd

from aatoolbox.config.countryconfig import get_country_config
from aatoolbox.datasources.codab.codab import CodAB


class Pipeline:
    """
    Initialize a configured pipeline.

    Parameters
    ----------
    iso3_unvalidated: str
        Country _iso3 passed in by the user
    """

    def __init__(self, iso3_unvalidated: str):
        self._config = get_country_config(iso3_unvalidated)
        self._codab = CodAB(self._config.iso3)

    def load_codab(
        self, admin_level: int, clobber: bool = False
    ) -> gpd.GeoDataFrame:
        """
        Get the COD AB data by admin level.

        Parameters
        ----------
        admin_level: int
            The administrative level
        clobber: bool
            If true, overwrite existing COD AB files

        Returns
        -------
        COD AB geodataframe with specified admin level

        Examples
        --------
        >>> from aatoolbox.pipeline import Pipeline
        >>> # Get admin 0 boundaries for Nepal
        >>> pipeline = Pipeline("npl")
        >>> npl_admin0 = pipeline.load_codab(admin_level=2)
        """
        admin_level_max = self._config.codab.admin_level_max
        if admin_level > admin_level_max:
            raise AttributeError(
                f"Admin level {admin_level} requested, but maximum set to "
                f"{admin_level_max} in {self._config.iso3.upper()} config file"
            )
        return self._load_codab(
            layer_name=self._config.codab.layer_base_name.format(
                admin_level=admin_level
            ),
            clobber=clobber,
        )

    def load_codab_custom(
        self, custom_layer_number: int = 0, clobber: bool = False
    ):
        """
        Get the COD AB data from a custom (non-level) layer.

        Parameters
        ----------
        custom_layer_number: int
            The 0-indexed number of the layer listed in the custom_layer_names
            parameter of the country's config file
        clobber: bool
            If true, overwrite existing COD AB files

        Returns
        -------
        COD AB geodataframe with custom admin level

        Examples
        --------
        >>> from aatoolbox.pipeline import Pipeline
        >>> # Get district boundaries for Nepal
        >>> pipeline = Pipeline("npl")
        >>> npl_admin0 = pipeline.load_codab_custom(custom_layer_number=0)
        """
        try:
            # Ignore mypy for this line because custom_layer_names could be
            # None, but this is handled by the caught exceptions
            layer_name = self._config.codab.custom_layer_names[
                custom_layer_number
            ]  # type: ignore
        except (IndexError, TypeError):
            raise AttributeError(
                f"{custom_layer_number}th custom layer requested but not "
                f"available in {self._config.iso3.upper()} config file"
            )
        return self._load_codab(layer_name=layer_name, clobber=clobber)

    def _load_codab(self, layer_name: str, clobber: bool):
        self._codab.download(
            hdx_address=self._config.codab.hdx_address,
            hdx_dataset_name=self._config.codab.hdx_dataset_name,
            clobber=clobber,
        )
        return self._codab.load_admin_layer(layer_name=layer_name)



"""Tests for the FewsNet module."""
import pytest

from aatoolbox.datasources.fewsnet.fewsnet import download_fewsnet


def test_download_fewsnet_country(mock_download_call, tmp_path):
    """Test that the correct country url and path is returned."""
    url, output_path = mock_download_call(
        date_pub="2020-10-01",
        iso2="et",
        region_name="east-africa",
        region_code="EA",
        country_data=True,
        region_data=False,
    )

    assert (
        url == "https://fdw.fews.net/api/ipcpackage/"
        "?country_code=ET&collection_date=2020-10-01"
    )

    assert output_path == tmp_path / "ET202010"


def test_download_fewsnet_region(mock_download_call, tmp_path):
    """Test that the correct region url and path is returned."""
    url, output_path = mock_download_call(
        date_pub="2020-10-01",
        iso2="et",
        region_name="east-africa",
        region_code="EA",
        country_data=False,
        region_data=True,
    )

    assert (
        url == "https://fews.net/data_portal_download/download?"
        "data_file_path=http://shapefiles.fews.net.s3.amazonaws.com/"
        "HFIC/EA/east-africa202010.zip"
    )

    assert output_path == tmp_path / "EA202010"


def test_download_fewsnet_nodata(mock_download_call):
    """Test that RuntimeError is returned when no data exists."""
    with pytest.raises(RuntimeError) as e:
        url, output_path = mock_download_call(
            date_pub="2020-10-01",
            iso2="et",
            region_name="east-africa",
            region_code="EA",
            country_data=False,
            region_data=False,
        )
        assert output_path is None
    assert "No data found for 2020-10" in str(e.value)


@pytest.fixture
def mock_download_call(mock_fake_url, mock_countryregion, tmp_path):
    """Mock call to download_fewsnet."""
    url_mock = mock_fake_url

    def _get_country_mock(
        date_pub, iso2, region_name, region_code, country_data, region_data
    ):
        mock_countryregion(country_data, region_data)

        output_path = download_fewsnet(
            date_pub=date_pub,
            iso2=iso2,
            region_name=region_name,
            region_code=region_code,
            output_dir=tmp_path,
            use_cache=False,
        )

        _, kwargs_download_url = url_mock.call_args
        url = kwargs_download_url["url"]

        return url, output_path

    return _get_country_mock


@pytest.fixture
def mock_fake_url(mocker):
    """Mock url and unzip call."""
    fakedownloadurl = mocker.patch(
        "aatoolbox.datasources.fewsnet.fewsnet.download_url"
    )
    mocker.patch("aatoolbox.datasources.fewsnet.fewsnet.unzip")
    return fakedownloadurl


@pytest.fixture
def mock_countryregion(mocker):
    """Mock that no country and/or region data exists."""

    def _mock_countryregion(country_data: bool, region_data: bool):
        if country_data:
            return None
        else:
            if region_data:
                return mocker.patch(
                    "aatoolbox.datasources.fewsnet.fewsnet."
                    "_download_fewsnet_country",
                    return_value=None,
                )
            else:
                return mocker.patch(
                    "aatoolbox.datasources.fewsnet.fewsnet."
                    "_download_fewsnet_country",
                    return_value=None,
                ), mocker.patch(
                    "aatoolbox.datasources.fewsnet.fewsnet."
                    "_download_fewsnet_region",
                    return_value=None,
                )

    return _mock_countryregion
